import ast
import re

from aws_access_util.constants import constants


def get_payload(form_response_soup, user_name, password):
    '''
    :param form_response_soup: The response returned after calling the url
    :param user_name: user name entered from cli
    :param password: password entered from cli
    :return: the payload for the next call
    '''
    payload = {}
    for inputtag in form_response_soup.find_all(re.compile('(INPUT|input)')):
        name = inputtag.get('name', '')
        value = inputtag.get('value', '')
        if "pf.username" in name.lower():
            # In the response this is the field for username
            payload[name] = user_name
        elif "email" in name.lower():
            # In the response this is the field for email, but this is not used
            payload[name] = user_name
        elif "pf.usertype" in name.lower():
            payload[name] = constants.user_type
        elif "pf.pass" in name.lower():
            # In the response this is the field for password
            payload[name] = password
        else:
            # Populate the parameter with the existing value (gets the hidden
            # fields in the login form)
            payload[name] = value
    return payload


def get_action_url(form_response_soup):
    '''
    :param form_response_soup: The response after calling the
    action url from the initial response
    :return: gets the auth form to be submitted next
    '''
    auth_form_submit_url = ''
    for input_tag in form_response_soup.find_all(re.compile('(FORM|form)')):
        action = input_tag.get('action')
        loginid = input_tag.get('id')
        if (action and loginid == "login-form"):
            auth_form_submit_url = action

    return auth_form_submit_url


def get_duo_attributes(form_response_soup):
    '''

    :param form_response_soup: response after submitting the auth url
    :return: the duo host name, signature and call back from duo init
    '''
    input_tag = form_response_soup.find_all(
        'script', {"type": "application/javascript"})
    duo_init = ast.literal_eval(
        input_tag[1].get_text().split(
            '(', 1)[1].split(')')[0])
    duo_host = duo_init['host']
    duo_signatures = duo_init['sig_request'].split(":")
    duo_callback = duo_init['post_action']

    return duo_host, duo_signatures, duo_callback


def get_duo_sid(form_response_soup):
    '''
    :param form_response_soup: response from auth url
    :return: gets the duo sid
    '''
    duo_sid = form_response_soup.find("input", {"name": "sid"})['value']
    return duo_sid


def get_assertion(form_response_soup):
    '''
    :param form_response_soup: response from the the final submission
    :return: sml assertion, this is what is used to connect to aws
    to get the temp token
    '''
    assertion = ''
    for input_tag in form_response_soup.find_all('input'):
        if (input_tag.get('name') == 'SAMLResponse'):
            # print(inputtag.get('value'))
            assertion = input_tag.get('value')
    return assertion



import base64
import configparser
import os
import sys
import traceback

import defusedxml.ElementTree as ET
from os.path import expanduser

import boto.s3
import boto.sts
from aws_access_util.constants import constants
import logging
from datetime import datetime


def saml2AWS(assertion, duration_seconds):
    if duration_seconds is None:
        duration_seconds = constants.default_access_duration

    logging.getLogger('boto').setLevel(logging.DEBUG)
    outputformat = constants.outputformat
    awsconfigfile = constants.awsconfigfile
    region = constants.region

    awsroles = []
    root = ET.fromstring(base64.b64decode(assertion))

    for saml2attribute in root.iter(
            '{urn:oasis:names:tc:SAML:2.0:assertion}Attribute'):
        if (saml2attribute.get('Name') ==
                'https://aws.amazon.com/SAML/Attributes/Role'):
            for saml2attributevalue in saml2attribute.iter(
                    '{urn:oasis:names:tc:SAML:2.0:assertion}AttributeValue'):
                awsroles.append(saml2attributevalue.text)

    # Note the format of the attribute value should be role_arn,principal_arn
    # but lots of blogs list it as principal_arn,role_arn so let's reverse
    # them if needed.
    for awsrole in awsroles:
        chunks = awsrole.split(',')
        if 'saml-provider' in chunks[0]:
            newawsrole = chunks[1] + ',' + chunks[0]
            index = awsroles.index(awsrole)
            awsroles.insert(index, newawsrole)
            awsroles.remove(awsrole)

    # If we have more than one role, ask the user which one they want,
    # otherwise just proceed.
    print("")
    if len(awsroles) > 1:
        i = 0
        print("Please choose the role you would like to assume:")
        for awsrole in awsroles:
            print('[', i, ']: ', awsrole.split(',')[0])
            i += 1

        try:
            print("Selection: ", end=' ')
            selectedroleindex = input()
        except KeyboardInterrupt:
            print('\n')
            print('###################################################################')
            print('# Process Interrupted...exiting out of the program                #')
            print('###################################################################')
            try:
                sys.exit(1)
            except SystemExit:
                os._exit(1)
        except Exception:
            traceback.print_exc(file=sys.stdout)

        # Basic sanity check of input
        if int(selectedroleindex) > (len(awsroles) - 1):
            print('You selected an invalid role index, please try again')
            sys.exit(0)

        role_arn = awsroles[int(selectedroleindex)].split(',')[0]
        principal_arn = awsroles[int(selectedroleindex)].split(',')[1]

    else:
        role_arn = awsroles[0].split(',')[0]
        principal_arn = awsroles[0].split(',')[1]

    # Use the assertion to get an AWS STS token using Assume Role with SAML
    conn = boto.sts.connect_to_region(region)
    try:
        token = conn.assume_role_with_saml(
            role_arn,
            principal_arn,
            assertion,
            duration_seconds=duration_seconds)
    except boto.exception.BotoServerError as e:
        track = traceback.format_exc()
        print(track)
        print('Failed to retrieve aws credentials : {e}'.format(e=e))
        print('###################################################################')
        print('#This may happen if you have specified an out of range duration   #')
        print('###################################################################')
        exit(1)
    except Exception as e:
        track = traceback.format_exc()
        print(track)
        print('Connection failure : {e}'.format(e=e))
        print('###################################################################')
        print('# Please connect with the support team and share the error message#')
        print('###################################################################')
        exit(1)
    # Write the AWS STS token into the AWS credential file
    home = expanduser("~")
    filename = home + awsconfigfile

    # Read in the existing config file
    config = configparser.RawConfigParser()
    config.read(filename)

    # Put the credentials into a saml specific section instead of clobbering
    # the default credentials
    if not config.has_section('saml'):
        config.add_section('saml')

    config.set('saml', 'output', outputformat)
    # config.set('saml', 'region', region)
    config.set('saml', 'aws_access_key_id', token.credentials.access_key)
    config.set('saml', 'aws_secret_access_key', token.credentials.secret_key)
    config.set('saml', 'aws_session_token', token.credentials.session_token)

    # Write the updated config file
    with open(filename, 'w+') as configfile:
        config.write(configfile)

    # Give the user some basic info as to what has just happened
    print('\n\n----------------------------------------------------------------')
    print(
        'Current time in UTC is {cur_date}'.format(
            cur_date=datetime.utcnow()))
    print('Your new access key pair has been stored in the AWS configuration '
          'file {0} under the saml profile.'.format(
              filename))
    print(
        'Note that it will expire at {0}.'.format(
            token.credentials.expiration))
    print('After this time, you may safely rerun this script to refresh '
          'your access key pair.')
    print(
        'To use this credential, call the AWS CLI with the '
        '--profile option (e.g. aws --profile saml ec2 describe-instances).')
    print('----------------------------------------------------------------\n\n')



# Note: Requires Python 3.3 or higher

# This module integrates with DUO MFA and generates a
# temporary access and token which are copied
# in the aws credentials file under ./aws. These
# credentials are then used to interact with AWS resources
import os.path
import getpass
import ast
import argparse
import time
import traceback
from os import path
from os.path import expanduser
import requests
from bs4 import BeautifulSoup
import sys
sys.path.append('..')
from aws_access_util.constants import constants # noqa: E402
from aws_access_util.constants.constants import call_back_url  # noqa: E402
from aws_access_util.constants.constants import awsconfigfile, \
    awsfolder  # noqa: E402
from jinja2 import Environment, FileSystemLoader, \
    BaseLoader  # noqa: E402
from aws_access_util.common.duo_utilities import get_payload, get_action_url, \
    get_duo_attributes, get_duo_sid, get_assertion  # noqa: E402
from aws_access_util.common.aws_utilities import saml2AWS  # noqa: E402
from aws_access_util.constants.constants import auth_url  # noqa: E402
from aws_access_util.constants.constants import prompt_url  # noqa: E402
from aws_access_util.constants.constants import status_url  # noqa: E402
from aws_access_util.constants.constants import result_url  # noqa: E402


def get_saml_assertion(url, user, password):
    """
    This function integrates with DUO mfa and returns the saml assertion
    :url : The URL of the AWS account
    :user: The userid to connect to the URL
    :password: The password to connect to AWS
    :return: saml assertion
    """
    cisco_url = url.split('/idp')[0]
    session = requests.Session()

    try:
        idp_url_response = session.get(url)

    except requests.exceptions.InvalidSchema as e:
        track = traceback.format_exc()
        print(track)
        print('Connection failure : {e}'.format(e=e))
        print('###################################################################')
        print('#This may happen if the URL is missing https:// at the start      #')
        print('###################################################################')
        exit(1)
    except requests.exceptions.MissingSchema as e:
        track = traceback.format_exc()
        print(track)
        print('Connection failure : {e}'.format(e=e))
        print('###################################################################')
        print('#This may happen if the URL provided is incorrect                 #')
        print('###################################################################')
        exit(1)
    except requests.exceptions.ConnectionError as e:
        track = traceback.format_exc()
        print(track)
        print('Connection failure : {e}'.format(e=e))
        print('###################################################################')
        print('#This may happen if the URL provided is incorrect or not reachable#')
        print('###################################################################')
        exit(1)
    except Exception as e:
        track = traceback.format_exc()
        print(track)
        print('Connection failure : {e}'.format(e=e))
        print('###################################################################')
        print('# Please connect with the support team and share the error message#')
        print('###################################################################')
        exit(1)

    idp_url_response_text = idp_url_response.text

    idp_url_response_soup = BeautifulSoup(
        idp_url_response_text, features="html.parser")
    payload = get_payload(idp_url_response_soup, user, password)
    del user
    del password

    action_url = get_action_url(idp_url_response_soup)

    # Invoke the action URL through a POST call.
    # This will return the duoinit which will need to be
    # parsed to get the duo host name, duo signatures
    # and the post action url
    try:
        response = session.post(
            action_url,
            data=payload,
            verify=constants.sslverification)
    except Exception as e:
        track = traceback.format_exc()
        print(track)
        print('Failed to get duoinit, error : {e}'.format(e=e))
        print('###################################################################')
        print('# Please connect with the support team and share the error message#')
        print('###################################################################')
        exit(1)
    response_text = response.text

    response_form_soup = BeautifulSoup(response_text, features="html.parser")
    # IF MFA is not required for generic accounts just return from here with
    # the assertion
    for inputtag in response_form_soup.find_all('input'):
        if (inputtag.get('name') == 'SAMLResponse'):
            # print(inputtag.get('value'))
            saml_assertion = inputtag.get('value')
            return saml_assertion

        else:
            break

    try:
        duo_host, duo_signatures, duo_callback = get_duo_attributes(
            response_form_soup)
    except IndexError as e:
        track = traceback.format_exc()
        print(track)
        print('Failed to get duo attribute, error : {e}'.format(e=e))
        print('###################################################################')
        print('# This may happen if you have not provided the right credentials  #')
        print('###################################################################')
        exit(1)
    except Exception as e:
        track = traceback.format_exc()
        print(track)
        print('Failed to get duo attribute, error : {e}'.format(e=e))
        print('###################################################################')
        print('# Please connect with the support team and share the error message#')
        print('###################################################################')
        exit(1)

    # Now call the auth URL to get the SID
    auth_url_formatted = auth_url.format(
        duo_host=duo_host) + "?tx=" + duo_signatures[0] \
        + "&parent=https://cloudsso.cisco.com/" \
        + duo_callback + "?&v=2.6"

    try:
        auth_response = session.post(auth_url_formatted)
    except Exception as e:
        track = traceback.format_exc()
        print(track)
        print('Failed to get response from auth url : {e}'.format(e=e))
        print('###################################################################')
        print('# Please connect with the support team and share the error message#')
        print('###################################################################')
        exit(1)

    auth_response_text = auth_response.text
    auth_response_form_soup = BeautifulSoup(
        auth_response_text, features="html.parser")
    duo_sid = get_duo_sid(auth_response_form_soup)
    prompt_url_formatted = prompt_url.format(duo_host=duo_host)

    # Prompts the user to enter the DUO authentication mechanism
    i = 0
    options = constants.options
    print("Choose an authentication method:")
    for option in options:
        print('[', i, ']: ', option.split(',')[0])
        i += 1
    try:
        print("Selection: ", end=' ')
        selectedauthenticationmethod = input()
    except KeyboardInterrupt:
        print('\n')
        print('###################################################################')
        print('# Process Interrupted...exiting out of the program                #')
        print('###################################################################')
        try:
            sys.exit(1)
        except SystemExit:
            os._exit(1)
    except Exception:
        traceback.print_exc(file=sys.stdout)

    factor = options[int(selectedauthenticationmethod)].split(',')[1]
    passcode = ''
    if (factor == 'sms' or factor == 'Passcode'):
        if(factor == 'sms'):
            prompt_url_formatted = prompt_url_formatted + "?sid=" + \
                duo_sid + "&device=phone1&factor=" + factor
            try:
                prompt_url_response = session.post(prompt_url_formatted)
                factor = 'Passcode'
            except Exception as e:
                track = traceback.format_exc()
                print(track)
                print('Failed to get response from sms : {e}'.format(e=e))
                print(
                    '###################################################################')
                print(
                    '# Please connect with the support team and share the error message#')
                print(
                    '###################################################################')
                exit(1)

        print("Enter Passcode: ", end=' ')
        try:
            passcode = input()
        except KeyboardInterrupt:
            print('\n')
            print('###################################################################')
            print('# Process Interrupted...exiting out of the program                #')
            print('###################################################################')
            try:
                sys.exit(1)
            except SystemExit:
                os._exit(1)
        except Exception:
            traceback.print_exc(file=sys.stdout)
        prompt_url_formatted = prompt_url_formatted + "?sid=" + duo_sid + \
            "&device=phone1&factor=" + factor + "&passcode=" + passcode
    else:
        prompt_url_formatted = prompt_url_formatted + "?sid=" + \
            duo_sid + "&device=phone1&factor=" + factor
    try:
        prompt_url_response = session.post(prompt_url_formatted)
    except Exception as e:
        track = traceback.format_exc()
        print(track)
        print('Failed to get response : {e}'.format(e=e))
        print('###################################################################')
        print('# Please connect with the support team and share the error message#')
        print('###################################################################')
        exit(1)

    prompt_url_response_text = prompt_url_response.text
    prompt_url_response_json = ast.literal_eval(prompt_url_response_text)
    duo_tx_stat = prompt_url_response_json["stat"]
    duo_tx_id = prompt_url_response_json["response"]["txid"]
    if duo_tx_stat != 'OK':
        print("Invalid response from duo mfa device")
        raise Exception

    # get the status of prompt
    status_url_formatted = status_url.format(duo_host=duo_host)
    status_url_formatted = status_url_formatted + \
        "?sid=" + duo_sid + "&txid=" + duo_tx_id
    while True:
        try:
            status_url_response = session.post(status_url_formatted)
        except Exception as e:
            track = traceback.format_exc()
            print(track)
            print('Failed to get response: {e}'.format(e=e))
            print('###################################################################')
            print('# Please connect with the support team and share the error message#')
            print('###################################################################')
            exit(1)

        status_url_response_text = status_url_response.text
        status_url_response_json = ast.literal_eval(status_url_response_text)

        try:
            duo_tx_result = status_url_response_json['response']['result']
            duo_tx_status = status_url_response_json['response']['status']
            duo_tx_status_code = \
                status_url_response_json['response']['status_code']
        except KeyError:
            duo_tx_result = ''

        try:
            duo_result_url = status_url_response_json['response']['result_url']
        except KeyError:
            duo_result_url = ''

        try:
            if duo_tx_result == 'FAILURE':
                raise Exception
        except Exception:
            print(duo_tx_status_code + "::" + duo_tx_status)
            exit(1)

        if duo_tx_result == 'SUCCESS':
            break
        else:
            time.sleep(0.2)

    # Get the cookie
    result_url_formatted = result_url.format(
        duo_host=duo_host, duo_result_url=duo_result_url)
    result_url_formatted = result_url_formatted + \
        "?sid=" + duo_sid + "&txid=" + duo_tx_id

    try:
        result_url_response = session.post(result_url_formatted)
    except Exception as e:
        track = traceback.format_exc()
        print(track)
        print('Failed to get response : {e}'.format(e=e))
        print('###################################################################')
        print('# Please connect with the support team and share the error message#')
        print('###################################################################')
        exit(1)

    result_url_response_text = result_url_response.text
    result_url_response_json = ast.literal_eval(result_url_response_text)
    sig_response = result_url_response_json['response']['cookie'] + \
        ":" + duo_signatures[1]

    # Call back parent URL with ths sig_response
    call_back_url_formatted = call_back_url.format(
        cisco_url=cisco_url,
        duo_callback=duo_callback,
        sig_response=sig_response)
    try:
        call_back_url_response = session.post(call_back_url_formatted)
    except Exception as e:
        track = traceback.format_exc()
        print(track)
        print('Failed to get response : {e}'.format(e=e))
        print('###################################################################')
        print('# Please connect with the support team and share the error message#')
        print('###################################################################')
        exit(1)
    call_back_url_response_text = call_back_url_response.text
    call_back_url_response_soup = BeautifulSoup(
        call_back_url_response_text, features="html.parser")
    saml_assertion = get_assertion(call_back_url_response_soup)

    return saml_assertion


def create_default_profile():
    """
    This function creates the default profile
    under .aws folder, if the profile does not
    exist already
    """
    home = expanduser("~")
    folder = home + awsfolder
    if not os.path.exists(folder):
        os.makedirs(folder)

    filename = home + awsconfigfile

    # If the credentials file exist under ./aws, then we do not need to create
    # a template
    if os.path.isfile(filename):
        return

    f = None
    bundle_dir = getattr(sys, '_MEIPASS', path.abspath(path.dirname(__file__)))
    template_path = path.join(bundle_dir, 'template')
    env = Environment(
        loader=FileSystemLoader(
            str(template_path)),
        autoescape=True)

    myString = "[default] \n" \
               "aws_access_key_id = \n" \
               "aws_secret_access_key = \n"
    template = Environment(
        loader=BaseLoader(),
        autoescape=True).from_string(myString)
    output_from_template = template.render()
    try:
        f = open(filename, "w+")
        f.write(output_from_template)
    except IOError:
        track = traceback.format_exc()
        print(track)
        print("File not accessible")
        print('###################################################################')
        print('# Please connect with the support team and share the error message#')
        print('###################################################################')
    finally:
        if f is not None:
            f.close()


def get_aws_access(url, user, password, duration_seconds):
    """
    this function has been created to call the aws access
    module in batch
    :url : The url of the AWS account
    :user: The user id to connect to AWS
    :password: The password to connect to AWS
    :duration_seconds: Duration for which access is required.
    The default duration is 1 hour
    """

    try:
        saml_assertion = get_saml_assertion(url, user, password)
        create_default_profile()
        saml2AWS(saml_assertion, duration_seconds)
        return 0
    except Exception as e:
        track = traceback.format_exc()
        print(track)
        print("File not accessible")
        print('###################################################################')
        print('# Please connect with the support team and share the error message#')
        print('###################################################################')
        return 1


if __name__ == "__main__":

    parser = argparse.ArgumentParser(
        description="Retrieves and stores a temporary token which is used to access AWS resource")
    parser.add_argument('--awsurl',
                        dest='url',
                        type=str,
                        help=("Specify a valid aws url"),
                        required=False)

    parser.add_argument('--u',
                        dest='user',
                        type=str,
                        help=("Specify your userid"),
                        required=False)

    parser.add_argument('--p',
                        dest='password',
                        type=str,
                        help=("Specify your password"),
                        required=False)

    parser.add_argument(
        '--dur',
        dest='duration_seconds',
        type=str,
        help=(
            "Specify duration in seconds for which you need access."
            "Must be > 900 and less than 43200. Default is 12 hour or 43200 secs."),
        required=False)

    args = parser.parse_args()

    url = args.url
    user = args.user
    duration_seconds = args.duration_seconds

    if args.url is None:
        try:
            print("AWS url:", end='')
            url = input()
        except KeyboardInterrupt:
            print('\n')
            print('###################################################################')
            print('# Process Interrupted...exiting out of the program                #')
            print('###################################################################')
            try:
                sys.exit(1)
            except SystemExit:
                os._exit(1)
        except Exception:
            traceback.print_exc(file=sys.stdout)
    else:
        url = args.url

    if args.user is None:
        try:
            print("Username:", end='')
            user = input()
        except KeyboardInterrupt:
            print('\n')
            print('###################################################################')
            print('# Process Interrupted...exiting out of the program                #')
            print('###################################################################')
            try:
                sys.exit(1)
            except SystemExit:
                os._exit(1)
        except Exception:
            traceback.print_exc(file=sys.stdout)
    else:
        user = args.user

    if args.password is None:
        try:
            password = getpass.getpass()
        except KeyboardInterrupt:
            print('\n')
            print('###################################################################')
            print('# Process Interrupted...exiting out of the program                #')
            print('###################################################################')
            try:
                sys.exit(1)
            except SystemExit:
                os._exit(1)
        except Exception:
            traceback.print_exc(file=sys.stdout)
    else:
        password = args.password

    try:
        get_aws_access(url, user, password, duration_seconds)
    except Exception:
        traceback.print_exc(file=sys.stdout)
    sys.exit(0)



# EXIT CODES
success = 0
failure = 1
# MESSAGES
support_contact = "Please connect with the support team"
# ALL DUO CONSTANTS
sslverification = True
auth_url = "https://{duo_host}/frame/web/v1/auth"
prompt_url = "https://{duo_host}/frame/prompt"
status_url = "https://{duo_host}/frame/status"
result_url = "https://{duo_host}{duo_result_url}"
call_back_url = "{cisco_url}{duo_callback}?sig_response={sig_response}"
user_type = 'cco'
# AWS Params
awsfolder = '/.aws/'
awsconfigfile = '/.aws/credentials'
outputformat = 'json'
region = 'us-west-2'
default_access_duration = 43200
# DUO OPTIONS
options = [
    'Send Me a Push,Duo+Push',
    'Call Me,Phone+Call',
    'SMS me a code,sms',
    'Passcode,Passcode']



################################################################################
#https://packaging.python.org/tutorials/packaging-projects/
#python setup.py sdist bdist_wheel
#python -m twine upload --repository-url https://test.pypi.org/legacy/ dist/*
#pip install --index-url https://test.pypi.org/simple/ --no-deps aws-access-rajdeb
#C:\aws-saml-access\token - this is where I have my PYPI token
###python -m twine upload dist/*
#python -m aws_access_util.get_aws_access
###############################################################################
import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="aws-access", # Replace with your own username
    version="1.0.9",
    author="Rajib Deb",
    author_email="rajdeb@cisco.com",
    description="This package is used to get temporary token for aws access",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://bitbucket-eng-sjc1.cisco.com/bitbucket/scm/con/hyperloop.git",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.7',
    install_requires=[
          'requests',
          'beautifulsoup4==4.8.2',
          'jinja2',
          'configparser',
          'boto',
           'boto3',
           'botocore',
           'awscli',
           'defusedxml'
      ]
)